package fr.amap.lidar.amapvox.commons;

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
import java.lang.reflect.Field;
import java.math.BigInteger;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import ucar.ma2.DataType;

/**
 *
 * @author Julien Heurtebize (julienhtbe@gmail.com)
 * @author Philippe Verley
 */
public class Voxel {

    /**
     * x-axis voxel centre coordinate
     */
    public int i;

    /**
     * y-axis voxel centre coordinate
     */
    public int j;
    /**
     * z-axis voxel centre coordinate
     */
    public int k;
    /**
     * number of shots crossing the voxel (sampling intensity)
     */
    public int nbSampling;
    /**
     * number of echoes
     */
    public int nbEchos;
    /**
     * cumulated ray length
     */
    public double lgTotal;
    /**
     * distance from voxel centre to ground level
     */
    public float ground_distance;
    /**
     * mean ray length (cumulated ray length / nbSampling)
     */
    public double lMeanTotal;
    /**
     * approximated transmittance
     */
    public double approxTransm;
    /**
     * numerically estimated transmittance
     */
    public double transmittance;
    /**
     * mean shot angle
     */
    public double angleMean;
    /**
     * cumulated entering beam surface
     */
    public double bsEntering;
    /**
     * cumulated intercepted beam surface
     */
    public double bsIntercepted;
    /**
     * cumulated potential beam surface
     */
    public double bsPotential;
    /**
     * plant area density
     */
    public double PadBVTotal;
    // variables for NAN correction algorithm
    public float passNumber;
    public float neighboursNumber;
    // transmittance
    public double[] trNumBsOut;
    // variable for numerical estimation of the transmittance
    public List<TrRecord> trRecords;
    public boolean fallbackTransm = false;
    
    /**
     * mean distance to laser
     */
    public double distLaser;
    /**
     * ray length standard deviation
     */
    public double sdLength;
    // variable for attenuation coefficient (F. Pimont's method)
    public Attenuation attenuationVar;
    /**
     * attenuation coefficient
     */
    public double attenuation;
    /**
     * attenuation bias correction factor
     */
    public double attenuation_biasCorr;
    // sampling intensity at sub voxel scale
    public BigInteger subSampling;
    /**
     * sub voxel exploration rate
     */
    public double explorationRate;

    public Voxel() {
    }

    public static enum MergeMode {

        SUM,
        NO_MERGE,
        CUSTOM,
        NAN,
        REMOVE,
        WEIGHTED_MEAN;
    }

    public enum OutputVariable {

        I_INDEX("i", "i index", DataType.INT, "scalar", true, MergeMode.NO_MERGE, true),
        J_INDEX("j", "j index", DataType.INT, "scalar", true, MergeMode.NO_MERGE, true),
        K_INDEX("k", "k index", DataType.INT, "scalar", true, MergeMode.NO_MERGE, true),
        PLANT_AREA_DENSITY("PadBVTotal", "Plant Area Density", DataType.FLOAT, "dimensionless", MergeMode.CUSTOM),
        MEAN_ANGLE("angleMean", "Mean angle", DataType.FLOAT, "degree", MergeMode.WEIGHTED_MEAN),
        ENTERING_BEAM_SURFACE("bsEntering", "Entering beam surface", DataType.FLOAT, "m2", MergeMode.SUM),
        INTERCEPTED_BEAM_SURFACE("bsIntercepted", "Intercepted beam surface", DataType.FLOAT, "m2", MergeMode.SUM),
        POTENTIAL_BEAM_SURFACE("bsPotential", "Potential beam surface", DataType.FLOAT, "m2", MergeMode.SUM),
        GROUND_DISTANCE("ground_distance", "Ground distance", DataType.FLOAT, "m", MergeMode.NO_MERGE),
        MEAN_TOTAL_LENGTH("lMeanTotal", "Mean ray length", DataType.FLOAT, "m", MergeMode.CUSTOM),
        TOTAL_LENGTH("lgTotal", "Total ray length", DataType.FLOAT, "m", MergeMode.SUM),
        NUMBER_OF_ECHOES("nbEchos", "Number of echoes", DataType.INT, "scalar", MergeMode.SUM),
        NUMBER_OF_SHOTS("nbSampling", "Number of shots", DataType.INT, "scalar", MergeMode.SUM),
        ESTIMATED_TRANSMITTANCE("transmittance", "Estimated transmittance", DataType.FLOAT, "dimensionless", false, MergeMode.CUSTOM, false),
        DIST_LASER("distLaser", "Mean distance to laser", DataType.FLOAT, "meter", MergeMode.WEIGHTED_MEAN),
        SD_LENGTH("sdLength", "Ray length standard deviation", DataType.FLOAT, "m2", false, MergeMode.WEIGHTED_MEAN, false),
        ATTENUATION_COEFF("attenuation", "Biased estimation of the attenuation coefficient", DataType.FLOAT, "dinensionless", MergeMode.WEIGHTED_MEAN),
        ATTENUATION_BIAS_CORR("attenuation_biasCorr", "Correction factor for the attenuation coefficient", DataType.FLOAT, "dinensionless", MergeMode.WEIGHTED_MEAN),
        EXPLORATION_RATE("explorationRate", "Subvoxel exploration rate", DataType.FLOAT, "dimensionless", false, MergeMode.CUSTOM, false),
        SUBSAMPLING("subSampling", "Sampling intensity at sub voxel scale", DataType.STRING, "binary", false, MergeMode.REMOVE, false);

        private final String shortName;
        private final String longName;
        private final DataType type;
        private final String units;
        private final boolean coordinateVariable;
        private final MergeMode mode;
        private final boolean enabledByDefault;

        private OutputVariable(String shortName, String longName, DataType type, String unit, boolean coordinateVariable, MergeMode mode, boolean enabledByDefault) {
            this.shortName = shortName;
            this.longName = longName;
            this.type = type;
            this.units = unit;
            this.coordinateVariable = coordinateVariable;
            this.mode = mode;
            this.enabledByDefault = enabledByDefault;
        }

        private OutputVariable(String shortName, String longName, DataType type, String unit, MergeMode mode) {
            this(shortName, longName, type, unit, false, mode, true);
        }

        public String getShortName() {
            return shortName;
        }

        /**
         * @return the longName
         */
        public String getLongName() {
            return longName;
        }

        /**
         * @return the type
         */
        public DataType getType() {
            return type;
        }

        /**
         * @return the unit
         */
        public String getUnits() {
            return units;
        }

        public boolean isCoordinateVariable() {
            return coordinateVariable;
        }

        public MergeMode getMode() {
            return mode;
        }

        public boolean isEnabledByDefault() {
            return enabledByDefault;
        }

        static public OutputVariable find(String variable) {

            for (OutputVariable ovar : OutputVariable.values()) {
                if (variable.trim().equalsIgnoreCase(ovar.shortName)) {
                    return ovar;
                }
            }
            throw new NullPointerException("Undefined output variable " + variable + ". Must be one of " + Arrays.toString(OutputVariable.values()));
        }
    }

    /**
     *
     * @param i
     * @param j
     * @param k
     */
    public Voxel(int i, int j, int k) {

        this.i = i;
        this.j = j;
        this.k = k;
        trRecords = new ArrayList();
        attenuationVar = new Attenuation();
    }

    public String variablesToString(String[] names, DecimalFormat dnf) {

        // voxel attributes to String
        StringBuilder voxelSB = new StringBuilder();
        for (String name : names) {
            try {
                Field field = Voxel.class.getField(name);
                switch (field.getType().getName()) {
                    case "double":
                    case "float":
                        double value = field.getDouble(this);
                        voxelSB.append(Double.isNaN(value) ? Double.NaN : dnf.format(value));
                        break;
                    case "int":
                        voxelSB.append(field.getInt(this));
                        break;
                    case "java.lang.String":
                        voxelSB.append((String) field.get(this));
                        break;
                    case "java.math.BigInteger":
                        voxelSB.append(((BigInteger) field.get(this)).toString(36));
                        break;
                }
            } catch (NoSuchFieldException | SecurityException | IllegalArgumentException | IllegalAccessException ex) {
                voxelSB.append(Double.NaN);
            }
            voxelSB.append(" ");
        }

        return voxelSB.toString().trim();
    }
    
    public void addTrRecord(double bfIn, double rayLength) {
        trRecords.add(new TrRecord(bfIn, rayLength));
    }

    public void setFieldValue(Class< ? extends Voxel> c, String fieldName, Object object, String value)
            throws NoSuchFieldException, SecurityException, IllegalArgumentException, IllegalAccessException {

        Field f = c.getField(fieldName);
        Class<?> type = f.getType();

        switch (type.getName()) {
            case "double":
                f.setDouble(object, Double.valueOf(value));
                break;
            case "float":
                f.setFloat(object, (float) Float.valueOf(value));
                break;
            case "int":
                f.setInt(object, Integer.valueOf(value));
                break;
            case "java.math.BigInteger":
                f.set(object, new BigInteger(value, 36));
        }
    }

    public double getFieldValue(Class< ? extends Voxel> c, String fieldName, Object o)
            throws SecurityException, NoSuchFieldException, IllegalAccessException {

        return c.getField(fieldName).getDouble(o);
    }

    public class TrRecord {

        public final double bsIn;
        public final double rayLength;

        private TrRecord(double bsIn, double rayLength) {
            this.bsIn = bsIn;
            this.rayLength = rayLength;
        }
    }

    public class Attenuation {

        public double interceptedBV;
        public double potentialBV;
        public double biasCorrNum;
    }
}
