/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.nephos3d.resources.programs;

import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.GL3bc;
import fr.amap.lidar.amapvox.nephos3d.resources.materials.Material;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import fr.amap.lidar.amapvox.nephos3d.resources.materials.StandardMaterial;
import fr.amap.lidar.amapvox.nephos3d.resources.shaders.Shader;
import fr.amap.lidar.amapvox.nephos3d.resources.textures.Texture2D;
import fr.amap.lidar.amapvox.nephos3d.resources.textures.TextureCubemap;
import java.util.List;
import net.jcip.annotations.GuardedBy;
import net.jcip.annotations.ThreadSafe;


/**
 * Default program to display simple models.
 */
@ThreadSafe
public class StandardProgram extends Program {

    @GuardedBy("this")
    private Texture2D albedo;
    
    @GuardedBy("this")
    private TextureCubemap cubemap;
    
    @GuardedBy("this")
    private boolean reflectionEnabled;
    
    @GuardedBy("this")
    private float[] mainColorValues;

    @GuardedBy("this")
    private int albedoLocation;
    
    @GuardedBy("this")
    private int cubemapLocation;
    
    @GuardedBy("this")
    private IntBuffer vbo;

    public StandardProgram() {
        super("StandardProgram");
        this.albedo = null;
        this.cubemap = null;
        this.reflectionEnabled = true;
        this.mainColorValues = new float[0];
        this.vbo = Buffers.newDirectIntBuffer(1);
        shader = new Shader("StandardShader");
        shader.setVertexSources(
                "/shaders/preprocessing/include/vertex.glsl",
                "/shaders/preprocessing/standard/vertex.glsl"
        );
        shader.setFragmentSources(
                "/shaders/preprocessing/include/fragment.glsl",
                "/shaders/preprocessing/standard/fragment.glsl"
        );
    }

    /**
     * Set if the reflexions are enabled.
     * @param reflectionEnabled 
     */
    public final void setReflectionEnabled(final boolean reflectionEnabled) {
        this.reflectionEnabled = reflectionEnabled;
    }

    @Override
    protected void onLoad(final GL3bc gl) {
        super.onLoad(gl);

        gl.glGenBuffers(vbo.limit(), vbo);

        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(0));
        gl.glEnableVertexAttribArray(6);
        gl.glVertexAttribPointer(6, 4, GL3bc.GL_FLOAT, false, 0, 0);
        gl.glVertexAttribDivisor(6, 1);

        albedoLocation = gl.glGetUniformLocation(shader.getProgram(), "albedo");
        gl.glUniform1i(albedoLocation, 0);
        cubemapLocation = gl.glGetUniformLocation(shader.getProgram(), "cubemap");
        gl.glUniform1i(cubemapLocation, 1);
    }

    @Override
    public void updateProperties() {
        super.updateProperties();
        
        try {
            final List<Material> materials = getMaterials();
            
            /* Get materials values */
            mainColorValues = new float[materials.size() * 4];

            for (int i = 0; i < materials.size(); i++) {
                final StandardMaterial material = (StandardMaterial) materials.get(i);

                if (material != null) {
                    if (material.isHiden()) {
                        mainColorValues[i * 4 + 0] = 1.0f;
                        mainColorValues[i * 4 + 1] = 1.0f;
                        mainColorValues[i * 4 + 2] = 1.0f;
                        mainColorValues[i * 4 + 3] = 0.0f;
                    } else {
                        mainColorValues[i * 4 + 0] = (float) 
                                material.getMainColor().getRed() / 255;
                        mainColorValues[i * 4 + 1] = (float) 
                                material.getMainColor().getGreen() / 255;
                        mainColorValues[i * 4 + 2] = (float) 
                                material.getMainColor().getBlue() / 255;
                        mainColorValues[i * 4 + 3] = (float) 
                                material.getMainColor().getAlpha() / 255;
                    }

                    cubemap = material.getCubemap();
                    albedo = material.getAlbedo();
                }
            }
        } catch(ArrayIndexOutOfBoundsException ex) {
            /* Ignore exception */
        }
    }

    @Override
    protected void onBind(final GL3bc gl) {
        super.onBind(gl);

        /* Enable model transformation matrix */
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(0));

        final FloatBuffer modelBuf
                = Buffers.newDirectFloatBuffer(mainColorValues);
        gl.glBufferData(GL3bc.GL_ARRAY_BUFFER, Float.BYTES * modelBuf.limit(),
                modelBuf, GL3bc.GL_STATIC_DRAW);

        /* Send albedo texture */
        gl.glActiveTexture(GL3bc.GL_TEXTURE0);

        if (albedo != null) {
            albedo.bind(gl);
        } else {
            gl.glBindTexture(GL3bc.GL_TEXTURE_2D, 0);
        }

        /* Send cubemap texture */
        gl.glActiveTexture(GL3bc.GL_TEXTURE1);

        if (cubemap != null && reflectionEnabled) {
            cubemap.bind(gl);
        } else {
            gl.glBindTexture(GL3bc.GL_TEXTURE_CUBE_MAP, 0);
        }
    }

}
